/*
  soundlow - soundcard-layer between ALSA driver and application

  soundlow by Robin Gruber <archi@robinsnest.de> (c) 2000
  Based on aplay by Jaroslav Kysela <perex@suse.cz>
  Based on vplay program by Michael Beck
  
*/


#include "soundlow.h"

#define DEBUG(x) 

SoundCard::SoundCard(int numberOfCards, int desiredRate, int desiredBufferSize)
{
    bufferSize = desiredBufferSize*sizeof(short)*2;
    cards=numberOfCards;
    rate=desiredRate;
    pcm_handle=(snd_pcm_t**)malloc(cards*sizeof(snd_pcm_t*));
    open();
}

SoundCard::~SoundCard()
{
    close();
}

void SoundCard::open()
{
    int card, dev,  err;
    char *cardname;
    struct snd_pcm_channel_params params;
    struct snd_pcm_channel_setup setup;
    dev = 0;   // use the first device of each card
    memset(&rformat, 0, sizeof(rformat));
    rformat.interleave = 1;
    rformat.format = SND_PCM_SFMT_S16;
    rformat.rate = rate;
    rformat.voices = 2;
    //  initialize all the cards
    for (card=0; card<cards; card++) {
        pcm_handle[card]=NULL;
        if ((err = snd_card_get_longname(card, &cardname)) < 0) {
            fprintf(stderr, "Error: unable to obtain longname: %s\n", snd_strerror(err));
            exit(1);
        }
        fprintf(stderr, "Using soundcard '%s'\n", cardname);
        free(cardname);
	
        if ((err = snd_pcm_open(&pcm_handle[card], card, dev, SND_PCM_OPEN_CAPTURE)) < 0) {
            fprintf(stderr, "Error: audio open error: %s\n", snd_strerror(err));
            exit(1);
        }
	format = rformat;
    }
    // the user wants a custom buffer size
    for (card=0; card<cards; card++) {
        snd_pcm_plugin_flush(pcm_handle[card], SND_PCM_CHANNEL_CAPTURE);  /* to be in right state */
        memset(&params, 0, sizeof(params));
        // prepare for reading data
        params.channel = SND_PCM_CHANNEL_CAPTURE; 
        memcpy(&params.format, &format, sizeof(format));

        // start captring on the go-command
        params.start_mode = SND_PCM_START_GO;
        params.stop_mode = SND_PCM_STOP_STOP;
        
        // switch to blocking mode = only get data when buffer is full
        params.mode = SND_PCM_MODE_BLOCK;
        params.buf.block.frag_size = bufferSize;
        params.buf.block.frags_max = -1;		/* little trick (playback only) */
        params.buf.block.frags_min = 1;

        // setup soundcard
        if (snd_pcm_plugin_params(pcm_handle[card], &params) < 0) {
            fprintf(stderr, "unable to set channel params\n");
            exit(1);
        }
        if (params.buf.block.frag_size!=bufferSize) {
            fprintf(stderr,"%d is not a valid buffer-size!\n",bufferSize);
            exit(1);
        }
        if (snd_pcm_plugin_prepare(pcm_handle[card], SND_PCM_CHANNEL_CAPTURE) < 0) {
            fprintf(stderr, " unable to prepare channel\n");
            exit(1);
        }
        memset(&setup, 0, sizeof(setup));
        setup.mode = SND_PCM_MODE_BLOCK;
        
        setup.channel = SND_PCM_CHANNEL_CAPTURE;
        if (snd_pcm_plugin_setup(pcm_handle[card], &setup) < 0) {
            fprintf(stderr, "unable to obtain setup\n");
            exit(1);
        }
    }
    DEBUG(fprintf(stderr,"SoundCard: real buffer_size = %i, frags = %i, total = %i\n", 
                  bufferSize, setup.buf.block.frags, 
                  setup.buf.block.frags * bufferSize));   
}


void SoundCard::close()
{
    // stop capturing 
    int card;
    for (card=0; card<cards; card++) {
        snd_pcm_close(pcm_handle[card]);
    }
}


void SoundCard::go(void)
{
    // start capturing
    int card;
    for (card=0; card<cards; card++) {
        snd_pcm_capture_go(pcm_handle[card]);
    }
}


bool SoundCard::captureReadError(int card)
{
    // some error has occured while capturing from card <card>
    // so let's see what we can do
    snd_pcm_channel_status_t status;
    memset(&status, 0, sizeof(status));
    status.channel = SND_PCM_CHANNEL_CAPTURE;
    fprintf(stderr,"SoundCard: Read error at card %d\n",card);
    // get the status information
    if (snd_pcm_plugin_status(pcm_handle[card], &status)<0) {
        fprintf(stderr, "SoundCard: capture channel status error\n");
        exit(1);
    }
    // everything is ok, but the driver is waiting for data 
    if (status.status == SND_PCM_STATUS_RUNNING)
        return false;	
    // the buffer is overrun, sync is lost!
    if (status.status == SND_PCM_STATUS_OVERRUN) {
        fprintf(stderr,"SoundCard: overrun at position %u!!!\n", status.scount);
        // first: reset the cards
        close();
        open();
        // second: start again
        go();
        // data should be accepted again!
        return true;		
    }
    fprintf(stderr, "SoundCard: read error\n");
    exit(1);
}


void SoundCard::read(signed short *data)
{
    int card;
    int l;
    // read data from the soundcard
    for (card=0; card<cards; card++) {
        do {
            l = snd_pcm_plugin_read(pcm_handle[card], (void*)&data[card*bufferSize/sizeof(short)], 
                                    bufferSize);
            DEBUG(fprintf(stderr," SoundCard: read: %1d: %5d\n",card,l));
            if (l == 0 || l == -EAGAIN) {
                usleep(10000);
                DEBUG(fprintf(stderr,"SoundCard: EAGAIN!\n"));
            }
            // if we get a buffer overrun, handle the error and restart
            // reading from the first card
            if (l == -EPIPE) {
                DEBUG(fprintf(stderr,"SoundCard: EPIPE\n"));
                if (captureReadError(card)){card=0;}
            }
            // if a non-fatal error occured, try again!
        } while ((l==0)||(l==-EAGAIN)||(l==-EPIPE));
        if (l < 0 && l != -EAGAIN) {
            fprintf(stderr, "SoundCard: read error: %s\n", snd_strerror(l));
            exit(-1);
        }
    }
}
